﻿using CelloSaaS.AuthorizationServer.Localization;
using CelloSaaS.AuthorizationServer.Models;
using CelloSaaS.AuthServer.Core.Ancilliary;
using CelloSaaS.AuthServer.Core.Extensions;
using CelloSaaS.AuthServer.Core.Models.Consents;
using CelloSaaS.AuthServer.Core.Results;
using CelloSaaS.AuthServer.Core.ServiceContracts;
using CelloSaaS.AuthServer.Core.SettingsManagement.Contracts;
using CelloSaaS.Library;
using CelloSaaS.ServiceContracts.TenantManagement;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Globalization;
using System.Linq;
using System.Threading.Tasks;
using System.Web.Http;
using System.Net.Http;

namespace CelloSaaS.AuthorizationServer.Controllers
{
    public class ErrorController : CelloApiController
    {
        [Route("signin-genericoauth2")]
        public IHttpActionResult Error()//string error, string error_description)
        {
            return null;
        }

    }

    [RoutePrefix(Routes.MetadataRoutePrefix)]
    public class MetadataController : CelloApiController
    {
        private readonly IConsentService _consentService;
        static string _logoImgFolderPath, _logoImgName;

        public MetadataController()
        {
            _consentService = ServiceLocator.Resolve<IConsentService>();

            if (string.IsNullOrEmpty(_logoImgFolderPath))
                _logoImgFolderPath = ConfigurationManager.AppSettings["LogoImageFolderPath"];

            if (string.IsNullOrEmpty(_logoImgName))
                _logoImgName = ConfigurationManager.AppSettings["DefaultLogo"];
        }

        #region Authentication Metadata
        [Route(Routes.GetAuthenticationProvidersRoute, Name = Routes.GetAuthenticationProvidersRouteName)]
        public async Task<IHttpActionResult> GetAuthenticationProviders()
        {
            try {
                Dictionary<string, AuthenticationProvider> authenticationTypes = await _consentService.GetAuthenticationProviders();
                if (authenticationTypes == null || authenticationTypes.Count < 1) {
                    return new ResponseContentResult<string>(AuthenticationResource.e_AuthenticationTypesFound);
                }

                return new ResponseContentResult<List<AuthenticationProvider>>(authenticationTypes.Values.ToList());
            }
            catch (SubjectDetailException exception) {
                Task.Run(() => ExceptionService.HandleException(exception, CelloSaaS.AuthServer.Core.LoggingConstants.AuthServerExceptionPolicy));
                return new ResponseContentResult<string>(AuthenticationResource.ex_AuthTypeFetchFailed);
            }
        }

        [Route(Routes.GetAuthenticationTypesRoute, Name = Routes.GetAuthenticationTypesRouteName)]
        public async Task<IHttpActionResult> GetAuthenticationTypesForProvider(string providerId = null)
        {
            try {
                Dictionary<string, AuthenticationType> authTypes = null;

                authTypes = (string.IsNullOrEmpty(providerId))
                            ? await _consentService.GetAuthenticationTypesAsync(null)
                            : await _consentService.GetAuthenticationTypesAsync(new string[] { providerId });

                Dictionary<string, Dictionary<string, string>> authTypesResult = new Dictionary<string, Dictionary<string, string>>(StringComparer.OrdinalIgnoreCase);

                foreach (var authType in authTypes) {
                    if (!authTypesResult.ContainsKey(authType.Value.ProviderId.ToString()))
                        authTypesResult.Add(authType.Value.ProviderId.ToString().ToLowerInvariant(), new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase));

                    if (!authTypesResult[authType.Value.ProviderId.ToString()].ContainsKey(authType.Value.Id.ToString()))
                        authTypesResult[authType.Value.ProviderId.ToString()].Add(authType.Value.Id.ToString().ToLowerInvariant(), authType.Value.Name);
                }
                return new ResponseContentResult<Dictionary<string, Dictionary<string, string>>>(authTypesResult);
            }
            catch (SubjectDetailException exception) {
                Task.Run(() => ExceptionService.HandleException(exception, CelloSaaS.AuthServer.Core.LoggingConstants.AuthServerExceptionPolicy));
                return new ResponseContentResult<string>(AuthenticationResource.ex_AuthTypeFetchFailed);
            }
        }
        #endregion

        #region Multi-Tenant Branding

        [HttpPost]
        [Route(Routes.CorporateLogoRoute, Name = Routes.CorporateLogoRouteName)]
        public string GetCorporateLogo([FromUri] string corpName)
        {
            if (string.IsNullOrEmpty(corpName)) return GetDefaultLogoUri();

            var tenantId = _tenantService.GetTenantIdFromTenantCode(corpName);

            if (string.IsNullOrEmpty(tenantId)) return GetDefaultLogoUri();

            var tenantSettings = _settingsService.Get(tenantId);

            if (tenantSettings == null || tenantSettings.Count < 1 || !tenantSettings.ContainsKey(CelloSaaS.ServiceContracts.SettingsManagement.AttributeConstants.Logo))
                return GetDefaultLogoUri();

            return string.Format(CultureInfo.InvariantCulture, "{0}/{1}/{2}", AppSettingHelpers.GetValue<string>(AppSettingConstants.ClientApplicationUri),
                _logoImgFolderPath, tenantSettings[CelloSaaS.ServiceContracts.SettingsManagement.AttributeConstants.Logo]);
        }

        private string GetDefaultLogoUri()
        {
            return string.Format(CultureInfo.InvariantCulture, "{0}/{1}/{2}", AppSettingHelpers.GetValue<string>(AppSettingConstants.ClientApplicationUri),
                _logoImgFolderPath, _logoImgName);
        }
        #endregion

        #region Federation Metadata Access


        public async Task<HttpResponseMessage> InitFederationMetadata(string federationMetadataUri, string companyCode, string clientId)
        {
            if (string.IsNullOrEmpty(federationMetadataUri))
                return this.Request.CreateErrorResponse(System.Net.HttpStatusCode.BadRequest, new ArgumentNullException(MetadataResource.ex_FederationMetadataUriNull));

            if (string.IsNullOrEmpty(companyCode))
                return this.Request.CreateErrorResponse(System.Net.HttpStatusCode.BadRequest, new ArgumentNullException(MetadataResource.ex_CompanyCodeNull));

            if (string.IsNullOrEmpty(clientId))
                return this.Request.CreateErrorResponse(System.Net.HttpStatusCode.BadRequest, new ArgumentNullException(MetadataResource.ex_ClientIdNull));

            var fedService = ServiceLocator.Resolve<CelloSaaS.AuthServer.Core.Contracts.Metadata.IFederationMetadataService>();

            try {
                var initStatus = fedService.Initialize(federationMetadataUri, companyCode, clientId);
                return this.Request.CreateResponse<bool>(initStatus);
            }
            catch (Exception exception) {
                ReportingService.Report("The following exception occured while initializing the federation metadata: " + exception.ToString());
                return this.Request.CreateErrorResponse(System.Net.HttpStatusCode.InternalServerError, exception);
            }
        }

        #endregion
    }

}
