﻿//-----------------------------------------------------------------------
// <copyright file="UserInfoController.cs" company="TechCello">
// Copyright (c) TechCello. All rights reserved.
// <author> TechCello </author>
// <date>8/24/2014 3:46:28 PM</date>
// </copyright>
//---------------------------------------------
using CelloSaaS.AuthorizationServer.Localization;
using CelloSaaS.AuthorizationServer.Models;
using CelloSaaS.AuthServer.Core;
using CelloSaaS.AuthServer.Core.Ancilliary;
using CelloSaaS.AuthServer.Core.Extensions;
using CelloSaaS.AuthServer.Core.Results;
using CelloSaaS.AuthServer.Core.Results.Errors;
using CelloSaaS.AuthServer.Core.ServiceContracts;
using CelloSaaS.AuthServer.Core.Services.Tokens;
using CelloSaaS.Library;
using CelloSaaS.Model.UserManagement;
using CelloSaaS.ServiceContracts.AccessControlManagement;
using CelloSaaS.ServiceContracts.UserManagement;
using System;
using System.Collections.Generic;
using System.IdentityModel.Tokens;
using System.Security.Claims;
using System.Threading.Tasks;
using System.Web.Http;

namespace CelloSaaS.AuthorizationServer.Controllers
{
    [RoutePrefix(Routes.UserInfoRoutePrefix)]
    public class UserInfoController : CelloApiController
    {
        private readonly IUserService _userService;
        private readonly IClientService _clientService;
        private readonly IRoleService _roleService;
        private readonly ITokenManager _tokenManager;
        private readonly IUserDetailsService _userDetailsService;

        public UserInfoController()
        {
            _userService = ServiceLocator.Resolve<IUserService>();
            _clientService = ServiceLocator.Resolve<IClientService>();
            _roleService = ServiceLocator.Resolve<IRoleService>();
            _tokenManager = ServiceLocator.Resolve<ITokenManager>();
            _userDetailsService = ServiceLocator.Resolve<IUserDetailsService>();
        }

        /// <summary>
        /// Returns the User Profile information 
        /// </summary>
        /// <returns></returns>
        public async Task<IHttpActionResult> Get(string client_id)
        {
            if (Request.Headers == null || Request.Headers.Authorization == null || string.IsNullOrEmpty(Request.Headers.Authorization.Parameter))
                return null;

            if (string.IsNullOrEmpty(client_id))
            {
                ReportingService.Report(AuthorizeResource.e_SecurityTokenValidationFailed);
                return new ErrorResult(AuthorizeResource.e_SecurityTokenValidationFailed);
            }

            // deconstruct the access token,
            string accessToken = Request.Headers.Authorization.Parameter;

            ClaimsPrincipal claimsPrincipal = null;
            try
            {
                claimsPrincipal = _tokenManager.ValidateToken(accessToken, client_id);
            }
            catch (SecurityTokenValidationException)
            {
                ReportingService.Report(AuthorizeResource.e_SecurityTokenValidationFailed);
                return new ErrorResult(AuthorizeResource.e_SecurityTokenValidationFailed);
            }

            // identify the client id and subject
            var clientClaim = claimsPrincipal.FindFirst(AuthServerConstants.ClaimTypes.ClientId);
            var userIdClaim = claimsPrincipal.FindFirst(System.Security.Claims.ClaimTypes.NameIdentifier);
            var scopeClaim = claimsPrincipal.FindFirst(AuthServerConstants.ClaimTypes.Scope);

            if ((clientClaim == null || string.IsNullOrEmpty(clientClaim.Value)) && (string.IsNullOrEmpty(client_id)))
            {
                ReportingService.Report(UserInfoResource.msg_clientnull);
                return Unauthorized(AuthenticationHeaderHelper.AuthenticateHeader);
            }

            if (userIdClaim == null || string.IsNullOrEmpty(userIdClaim.Value))
                userIdClaim = claimsPrincipal.FindFirst(AuthServerConstants.ClaimTypes.Subject);

            if (userIdClaim == null || string.IsNullOrEmpty(userIdClaim.Value))
            {
                ReportingService.Report(UserInfoResource.msg_useridclaimnull);
                return Unauthorized(AuthenticationHeaderHelper.AuthenticateHeader);
            }

            // get the client and get its tenantid
            var client = await _clientService.GetAsync(clientClaim != null ? clientClaim.Value : client_id);

            if (client == null)
            {
                ReportingService.Report(UserInfoResource.msg_clientnull);
                return Unauthorized(AuthenticationHeaderHelper.AuthenticateHeader);
            }

            return await PrepareUserProfile(userIdClaim.Value, scopeClaim != null ? scopeClaim.Value : string.Empty, client.Id.ToString());
        }

        /// <summary>
        /// Gets the profile based on the company code and the client identifier
        /// </summary>
        /// <param name="company_code">The company code</param>
        /// <param name="client_id">the client identifier</param>
        /// <returns></returns>
        public async Task<IHttpActionResult> GetProfile(string company_code, string client_id)
        {
            if (Request.Headers == null || Request.Headers.Authorization == null || string.IsNullOrEmpty(Request.Headers.Authorization.Parameter))
                return null;

            if (string.IsNullOrEmpty(client_id))
            {
                ReportingService.Report(AuthorizeResource.e_SecurityTokenValidationFailed);
                return new ErrorResult(AuthorizeResource.e_SecurityTokenValidationFailed);
            }

            //TODO: think about making this better
            //if(!string.IsNullOrEmpty(company_code))
            //{
            //    string tenantId = ServiceLocator.Resolve<CelloSaaS.ServiceContracts.TenantManagement.ITenantService>().GetTenantIdFromTenantCode(company_code);

            //    if(!string.IsNullOrEmpty(tenantId))
            //    {
            //        var tenantSettings = ServiceLocator.Resolve<CelloSaaS.AuthServer.Core.SettingsManagement.Contracts.IClientSettingsService>().Get(tenantId);

            //        if (tenantSettings != null && tenantSettings.Count > 0 && tenantSettings.ContainsKey(CelloSaaS.AuthServer.Core.Models.SettingAttributeConstants.IdentityProviders))
            //        {
            //            if(!tenantSettings[CelloSaaS.AuthServer.Core.Models.SettingAttributeConstants.IdentityProviders])
            //        }
            //    }
            //}

            // deconstruct the access token,
            string accessToken = Request.Headers.Authorization.Parameter;

            ClaimsPrincipal claimsPrincipal = null;
            try
            {
                //company code, access token & client id
                claimsPrincipal = _tokenManager.ValidateOpenIdConnectToken(accessToken, company_code, client_id);
            }
            catch (SecurityTokenValidationException)
            {
                ReportingService.Report(AuthorizeResource.e_SecurityTokenValidationFailed);
                return new ErrorResult(AuthorizeResource.e_SecurityTokenValidationFailed);
            }

            var upnClaim = claimsPrincipal.FindFirst(ClaimTypes.Upn);

            if (upnClaim == null)
                upnClaim = claimsPrincipal.FindFirst(ClaimTypes.Name);

            if (upnClaim == null || string.IsNullOrEmpty(upnClaim.Value))
            {
                ReportingService.Report("Cannot be able to find the user name from the UPN claim");
                return null;
            }

            string emailId = null, username = upnClaim.Value, storeUserId = null;

            var emailClaim = claimsPrincipal.FindFirst(ClaimTypes.Email);
            if (emailClaim != null && string.IsNullOrEmpty(emailClaim.Value))
                emailId = emailClaim.Value;

            return await PrepareUserProfile(null, null, client_id, username, emailId, claimsPrincipal);
        }

        /// <summary>
        /// Performs new user Registration with the Cello Membership Provider
        /// </summary>
        /// <returns></returns>
        public async Task<IHttpActionResult> PutAsync()
        {
            return null;
        }

        /// <summary>
        /// Update the user details with the Cello Membership Provider
        /// </summary>
        /// <returns></returns>
        public async Task<IHttpActionResult> PostAsync()
        {
            return null;
        }

        #region Private Members

        private async Task<IHttpActionResult> PrepareUserProfile(string userId, string scope, string clientId, string userName = null, string emailId = null, ClaimsPrincipal principal = null)
        {
            Tuple<UserDetails, List<string>> userProfile = await TryGetUserProfile(userId, userName, emailId);

            // if none of these help, log and fail
            if (userProfile == null || userProfile.Item1 == null)
            {
                //ReportingService.Report("unable to get the user profile based on the access token information");
                //return Unauthorized(AuthenticationHeaderHelper.AuthenticateHeader);

                Claim useridClaim = principal.FindFirst(ClaimTypes.NameIdentifier);

                return new ProfileResult(
                new
                {
                    id = userId ?? ((useridClaim != null) ? useridClaim.Value : string.Empty),
                    displayName = userName,
                    email = emailId ?? userName,
                    client_id = clientId,
                    scope = scope
                });
            }

            UserDetails profileInfo = userProfile.Item1;

            if (profileInfo == null || profileInfo.MembershipDetails == null || string.IsNullOrEmpty(profileInfo.MembershipDetails.TenantCode))
            {
                ReportingService.Report(UserInfoResource.msg_profilenull);
                return Unauthorized(AuthenticationHeaderHelper.AuthenticateHeader);
            }

            userId = string.IsNullOrEmpty(userId) ? profileInfo.User.Identifier : userId;

            // get the user roles
            string[] userroles = (userProfile.Item2 != null && userProfile.Item2.Count > 0)
                                ? userProfile.Item2.ToArray()
                                : _roleService.GetUserRoles(userId, profileInfo.MembershipDetails.TenantCode);

            // return the user details as a profile object
            return new ProfileResult(
                new
                {
                    id = profileInfo.User.Identifier,
                    displayName = profileInfo.MembershipDetails.UserName,
                    givenName = profileInfo.User.FirstName,
                    familyName = profileInfo.User.LastName,
                    email = profileInfo.MembershipDetails.EmailId,
                    tenantId = profileInfo.MembershipDetails.TenantCode,
                    rights = (userroles != null && userroles.Length > 0) ? string.Join(",", userroles) : string.Empty,
                    client_id = clientId,
                    scope = scope
                });
        }

        private async Task<Tuple<UserDetails, List<string>>> TryGetUserProfile(string userId, string userName, string emailId)
        {
            Tuple<UserDetails, List<string>> userProfile = null;
            if (!string.IsNullOrEmpty(userId))
            {
                // get the userdetails by userid
                userProfile = await _userService.Get(userId);
            }

            //1. Try to get the userdetails by email id
            if (userProfile == null && !string.IsNullOrEmpty(emailId))
            {
                UserDetails userDetails = _userDetailsService.GetUserDetailsByEmailId(emailId);

                if (userDetails != null && userDetails.MembershipDetails != null)
                    userProfile = new Tuple<UserDetails, List<string>>(userDetails, null);
            }

            //2. if not able to get, try getting by the username
            if (userProfile == null && !string.IsNullOrEmpty(userName))
            {
                //1. Try to get the userdetails by email id
                UserDetails userDetails = _userDetailsService.GetUserDetailsByUserName(userName);

                if (userDetails != null && userDetails.MembershipDetails != null)
                    userProfile = new Tuple<UserDetails, List<string>>(userDetails, null);
            }
            return userProfile;
        }

        #endregion

    }
}
